/*
 * Copyright@ Samsung Electronics Co. LTD
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied or distributed, transmitted,
 * transcribed, stored in a retrieval system or translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed
 * to third parties without the express written permission of Samsung Electronics.


 * Alternatively, this program is free software in case of open source projec;
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.

*/

#include "bl2_config.h"
#include "board_config.h"

/* cpu_state flag */
#define RESET			(1 << 0)
#define RESERVED		(1 << 1)
#define HOTPLUG			(1 << 2)
#define C2_STATE		(1 << 3)
#define CLUSTER_OFF		(1 << 4)

#define DFD_EDPCSR_DUMP_EN	(1 << 0)

.macro	push, xreg1, xreg2
stp	\xreg1, \xreg2, [sp, #-16]!
.endm

.macro	pop, xreg1, xreg2
ldp	\xreg1, \xreg2, [sp], #16
.endm

.global _start
_start:
	b	bl2_start

/*
 * Main code of bl2 (Cold boot and Warm boot)
 */
bl2_start:
	/*
	 * Could be EL3/EL2/EL1, Initial State:
	 * Little Endian, MMU Disabled, i/dCache Disabled
	 */
#ifdef CONFIG_ENABLE_EXCEPTION_VECTOR
	adr	x0, vectors
#endif
	mrs	x1, CurrentEL
	cmp x1, 0x8
	beq __start_el2
	cmp x1, 0x4
	beq __start_el1
__start_el3:
#ifdef CONFIG_ENABLE_EXCEPTION_VECTOR
	msr	vbar_el3, x0
#endif
	mrs	x0, scr_el3
	orr	x0, x0, #0xf			/* SCR_EL3.NS|IRQ|FIQ|EA */
	msr	scr_el3, x0
	b	0f
__start_el2:
#ifdef CONFIG_ENABLE_EXCEPTION_VECTOR
	msr	vbar_el2, x0
#endif
	b	0f
__start_el1:
#ifdef CONFIG_ENABLE_EXCEPTION_VECTOR
	msr	vbar_el1, x0
#endif
	mov x0, #3 << 20
	msr cpacr_el1, x0

0:
	/* use iRAM stack in bl2 */
	ldr	x0, =(CONFIG_IRAM_STACK)
	bic	sp, x0, #0xf

	/* check warm reset status */
	bl	check_warm_reset

	/* check reset status */
	ldr	x0, =(EXYNOS7570_POWER_INFORM1)
	ldr	w1, [x0]

	/* Sleep wakeup reset */
	ldr	w2, =S5P_CHECK_SLEEP
	cmp	w1, w2
	beq	wakeup_reset

	/* PS-Hold high */
	ldr	x0, =(EXYNOS7570_POWER_PS_HOLD_CONTROL)
	ldr	w1, =0x5300
	str	w1, [x0]

	bl	init_chip_info
	bl	pmic_enable_peric_dev

	bl	ect_init

	/* if watchdog or warm && dump GPR */
	ldr	x1, =(EXYNOS7570_POWER_DFD_RESET_SEQUENCER_CONFIGURATION)
	ldr	w2, [x1]
	and	w2, w2, #(DFD_EDPCSR_DUMP_EN)
	cmp	w2, #0
	b.eq	skip_dump_gpr

	/* check reset status */
	ldr	x1, =(EXYNOS7570_POWER_RST_STAT)
	ldr	w2, [x1]
	ldr	w3, =(WRESET | APOLLO_WDTRESET | ATLAS_WDTRESET)
	and	w2, w2, w3
	cmp	w2, #0
	b.ne	run_dump_gpr

skip_dump_gpr:
	bl	set_cpu_state

	/* relocate nscode to non-secure region */
	bl	relocate_code

run_dump_gpr:
	bl	pmic_init

	/* init system clock */
	bl	cmu_dmc_init

	bl	mark_bl2

	bl	load_ect_to_dram

	/* load binaries from storage using smc calls. */
	b	bl2_load_binary

/*****************************************************************************/
/* End of cold boot sequence.                                                */
/*****************************************************************************/
/*****************************************************************************/
/* Start of wakeup sequence.                                                 */
/*****************************************************************************/
wakeup_reset:

	bl	set_cpu_state

	/* relocate nscode to non-secure region */
	bl	relocate_code

	/* init system clock */
	bl	cmu_dmc_init

	bl	start_mct_frc

	/* If eMMC booting */
	/*
	ldr	x0, =(EXYNOS7570_POWER_INFORM3)
	ldr	w1, [x0]
	cmp	w1, #BOOT_EMMC_4_4
	bne	1f
	*/

	bl	msh2_fifo_reset
1:
	b	warmboot

/*****************************************************************************/
/* End of wakeup sequence.                                                   */
/*****************************************************************************/

set_cpu_state:
	/* read boot cluster */
	adr	x0, _cpu_state
	mrs	x1, mpidr_el1		/* read MPIDR */
	ubfx	w1, w1, #8, #4		/* w1 = cluster id */
	lsl	x1, x1, #4

	/* set reset flag at _cpu_state of boot cpu */
	ldr	w2, =RESET
	str	w2, [x0, x1]

	ret

check_warm_reset:
	/* check reset status */
	ldr	x0, =(EXYNOS7570_POWER_RST_STAT)
	ldr	w1, [x0]
	and	w1, w1, #WRESET
	tst	w1, #WRESET		/* check warm reset */
	beq	1f
	/* clear INFORM1 */
	ldr	x0, =(EXYNOS7570_POWER_INFORM1)
	mov	w1, #0x0
	str	w1, [x0]
1:
	ret

msh2_fifo_reset:
	/* Clear FIFO */
	ldr	x0, =EXYNOS7570_MMC0_BASE
	ldr	w1, [x0]
	orr	w1, w1, #0x2
	str	w1, [x0]

	ret

start_mct_frc:
	ldr	x0, =(EXYNOS7570_MCT_BASE + EXYNOS7570_G_TCON)
	ldr	w1, [x0]
	orr	w1, w1, #0x100	/* Timer Enable */
	str	w1, [x0]

	ret

/*
 * Relocate code
 */
relocate_code:
	adr	x0, nscode_base			/* r0: source address (start) */
	adr	x1, nscode_end			/* r1: source address (end) */
	ldr	x2, =CONFIG_PHY_IRAM_NS_BASE	/* r2: target address */

1:
	ldp	x3, x4, [x0], #16
	stp	x3, x4, [x2], #16
	cmp	x0, x1
	b.lt	1b

	isb	sy

	ret

load_ect_to_dram:
	/* start */
	ldr	x0, =(CONFIG_SPL_TEXT_BASE+PART_SIZE_BL2_CODE + PART_SIZE_APM)
	/* end */
	ldr	x1, =(CONFIG_SPL_TEXT_BASE+PART_SIZE_BL2_CODE + PART_SIZE_APM + PART_SIZE_ECT)
	/* target address */
	ldr	x2, =(CONFIG_SYS_ECT_BASE)

1:
	ldp	x3, x4, [x0], #16
	stp	x3, x4, [x2], #16
	cmp	x0, x1
	b.lt	1b

	dsb	sy
	isb	sy

	ret

/******************************************************************************/
/*
 * CPU1, 2, 3, 4, 5 waits here until CPU0 wake it up.
 * - below code is copied to CONFIG_PHY_IRAM_NS_BASE, which is non-secure memory.
 */
	.align 3
nscode_base:
	b	1f
	nop				/* for backward compatibility */

	.align 3
	.word	0x0			/* REG0: RESUME_ADDR */
	.word	0x0			/* REG1: RESUME_FLAG */
	.word	0x0			/* REG2 */
	.word	0x0			/* REG3 */
_switch_addr:
	.word	0x0			/* REG4: SWITCH_ADDR */
_hotplug_addr:
	.word	0x0			/* REG5: CPU1_BOOT_REG */
	.word	0x0			/* REG6 */
_c2_addr:
	.word	0x0			/* REG7: REG_C2_ADDR */
_cpu_state:
	.word	HOTPLUG			/* CLUSTER0_CORE0_STATE */
	.word	HOTPLUG			/* CLUSTER0_CORE1_STATE */
	.word	HOTPLUG			/* CLUSTER0_CORE2_STATE */
	.word	HOTPLUG			/* CLUSTER0_CORE3_STATE */
	.word	HOTPLUG			/* CLUSTER1_CORE0_STATE */
	.word	HOTPLUG			/* CLUSTER1_CORE1_STATE */
	.word	HOTPLUG			/* CLUSTER1_CORE2_STATE */
	.word	HOTPLUG			/* CLUSTER1_CORE3_STATE */
	.word	0x0			/* CLUSTER0_STATE */
	.word	0x0			/* CLUSTER1_STATE */
_lpc_state:
	.word	0x0			/* LPC FLAG */
	.word	0x0			/* LPC STATE */

1:
	adr	x0, _cpu_state

	mrs	x7, mpidr_el1
	ubfx	w8, w7, #8, #4		/* w8 = cluster id */
	and	w7, w7, #0xf		/* w7 = cpu id */
	lsl	w8, w8, #2
	add	w7, w7, w8

	/* read the current cpu state */
	ldr	w10, [x0, x7, lsl #2]

	/* clear cpu state */
	mov	x1, xzr
	str	w1, [x0, x7, lsl #2]

ns_svc_entry:
	nop
	mov	x0, #C2_STATE
	orr	x0, x0, #CLUSTER_OFF
	tst	x10, x0
	b.eq	2f
	adr	x0, _c2_addr
	b	wait_for_addr
2:
	/* clear INFORM1 for security reason  */
	ldr	x0, =(EXYNOS7570_POWER_INFORM1)
	ldr	w1, [x0]
	cmp	w1, #0x0
	b.eq	3f
	mov	w1, #0x0
	str	w1, [x0]
	ldr	x1, =(EXYNOS7570_POWER_INFORM0)
	ldr	w2, [x1]
	br	x2
3:
	tst	x10, #RESET
	b.eq	4f
	ldr	x1, =CONFIG_SYS_TEXT_BASE
	br	x1
4:
	adr	x0, _hotplug_addr
wait_for_addr:
	ldr	w1, [x0]
	cmp	w1, #0x0
	b.eq	5f
	br	x1
5:
	b	wait_for_addr
	.ltorg
nscode_end:

#if defined(CONFIG_ENABLE_EXCEPTION_VECTOR)
.section	".vector_handler",	"ax"
/*
 * Exception vectors.
 */
	.align 11
vectors:
	.align 7
bad_sync:
	b	.		/* Current EL Synchronous Thread */
	.align 7
bad_irq:
	b	.		/* Current EL IRQ Thread */
	.align 7
bad_fiq:
	b	.		/* Current EL FIQ Thread */
	.align 7
bad_error:
	b	.		/* Current EL Error Thread */
	.align 7
do_sync:
	b	.		/* Current EL Synchronous Handler */
	.align 7
do_irq:
	b	.		/* Current EL IRQ Handler */
	.align 7
do_fiq:
	b	.		/* Current EL FIQ Handler */
	.align 7
do_error:
	b	.		/* Current EL Error Handler */
	.align	7
lower_sync:
	b	.		/* Lower EL Synchronous for aarch64 */
	.align	7
lower_irq:
	b	.		/* Lower EL IRQ for aarch64 */
	.align	7
lower_fig:
	b	.		/* Lower EL FIQ for aarch64 */
	.align	7
lower_error:
	b	.		/* Lower EL Error for aarch64 */
#endif
