/* drivers/video/exynos/7880/decon_reg_7880.c
 *
 * CAL file for Samsung DECON of EXYNOS7880.
 *
 * Copyright@ Samsung Electronics Co. LTD
 *
 * Manseok Kim <manseoks.kim@samsung.com> *
 * Seuni Park <seuni.park@samsung.com> *
 *
 * This software is proprietary of Samsung Electronics.
 * No part of this software, either material or conceptual may be copied or distributed, transmitted,
 * transcribed, stored in a retrieval system or translated into any human or computer language in any form by any means,
 * electronic, mechanical, manual or otherwise, or disclosed
 * to third parties without the express written permission of Samsung Electronics.


 * Alternatively, this program is free software in case of open source projec;
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.

*/


/* use this definition when you test CAL on firmware */
/* #define FW_TEST */
#ifdef FW_TEST
#include "decon_fw.h"
#define __iomem
#else
#include "../decon.h"
#endif

/******************* CAL raw functions implementation *************************/
int decon_reg_reset(u32 id)
{
	int tries;

	decon_write(id, GLOBAL_CONTROL, GLOBAL_CONTROL_SRESET);
	for (tries = 2000; tries; --tries) {
		if (~decon_read(id, GLOBAL_CONTROL) & GLOBAL_CONTROL_SRESET)
			break;
		udelay(10);
	}

	if (!tries) {
		decon_err("failed to reset Decon\n");
		return -EBUSY;
	}

	return 0;
}

void decon_reg_set_clkgate_mode(u32 id, u32 en)
{
	u32 val = en ? ~0 : 0;

	if (id == 0)
		decon_write_mask(id, CLOCK_CONTROL_0, val, CLOCK_CONTROL_0_F_MASK);
}

void decon_reg_set_vclk_freerun(u32 id, u32 en)
{
	u32 val = en ? ~0 : 0;

	if (id == 0)
		decon_write_mask(0, DISPIF0_DISPIF1_CONTROL, val, DISPIF0_DISPIF1_CONTROL_FREE_RUN_EN);
}

/*
 * VCLK is the video source clock of DECON
 * NUM   : NUM_VALUE_OF_VCLK[7:0]
 * DENOM : DENOM_VALUE_OF_VCLK[23:16]
*/
void decon_reg_set_vclk_divider(u32 id, u32 denom, u32 num)
{
	u32 val;

	if (num == 1)
		denom = (denom * 2) - 1;
	else {
		num = num - 1;
		denom = denom - 1;
	}

	val = DIVIDER_DENOM_VALUE_OF_CLK_F(denom)
		| DIVIDER_NUM_VALUE_OF_CLK_F(num);

	/* VCLK controlled by DECON_F(id=0) only */
	if (id == 0)
		decon_write(0, VCLK_DIVIDER_CONTROL, val);
}

void decon_reg_set_data_path(u32 id, enum decon_data_path data_path, enum decon_enhance_path enhance_path)
{
	u32 val, mask;

	val = DATA_PATH_CONTROL_ENHANCE(enhance_path) | DATA_PATH_CONTROL_PATH(data_path);
	mask = DATA_PATH_CONTROL_ENHANCE_MASK | DATA_PATH_CONTROL_PATH_MASK;

	decon_write_mask(id, DATA_PATH_CONTROL, val, mask);
}

void decon_reg_get_data_path(u32 id, enum decon_data_path *data_path, enum decon_enhance_path *enhance_path)
{
	u32 val;

	val = decon_read(id, DATA_PATH_CONTROL);

	*data_path = DATA_PATH_CONTROL_PATH_GET(val);
	*enhance_path = DATA_PATH_CONTROL_ENHANCE_GET(val);
}

void decon_reg_set_operation_mode(u32 id, enum decon_psr_mode mode)
{
	if (mode == DECON_MIPI_COMMAND_MODE)
		decon_write_mask(id, GLOBAL_CONTROL, GLOBAL_CONTROL_OPERATION_MODE_I80IF_F,
						 GLOBAL_CONTROL_OPERATION_MODE_F);
	else
		decon_write_mask(id, GLOBAL_CONTROL, GLOBAL_CONTROL_OPERATION_MODE_RGBIF_F,
						 GLOBAL_CONTROL_OPERATION_MODE_F);
}

void decon_reg_set_blender_bg_image_size(u32 id, enum decon_dsi_mode dsi_mode, struct decon_lcd *lcd_info)
{
	u32 val, mask;

	val = BLENDER_BG_HEIGHT_F(lcd_info->yres) | BLENDER_BG_WIDTH_F(lcd_info->xres);
	mask = BLENDER_BG_HEIGHT_MASK | BLENDER_BG_WIDTH_MASK;
	decon_write_mask(id, BLENDER_BG_IMAGE_SIZE_0, val, mask);

	val = (lcd_info->yres) * (lcd_info->xres);
	decon_write_mask(id, BLENDER_BG_IMAGE_SIZE_1, val, ~0);
}

void decon_reg_get_blender_bg_image_size(u32 id, u32 *p_width, u32 *p_height)
{
	u32 val;

	val = decon_read(id, BLENDER_BG_IMAGE_SIZE_0);

	*p_width = BLENDER_BG_WIDTH_GET(val);
	*p_height = BLENDER_BG_HEIGHT_GET(val);
}

void decon_reg_set_underrun_scheme(u32 id, enum decon_hold_scheme mode)
{
	if (id == 0)
		decon_write_mask(0, DISPIF0_DISPIF1_CONTROL, DISPIF0_DISPIF1_CONTROL_UNDERRUN_SCHEME_F(mode), DISPIF0_DISPIF1_CONTROL_UNDERRUN_SCHEME_MASK);
}

void decon_reg_set_rgb_order(u32 id, enum decon_rgb_order order)
{
	if (id == 0)
		decon_write_mask(0, DISPIF0_DISPIF1_CONTROL, DISPIF0_DISPIF1_CONTROL_OUT_RGB_ORDER_F(order), DISPIF0_DISPIF1_CONTROL_OUT_RGB_ORDER_MASK);
}

void decon_reg_set_frame_fifo_size(u32 id, enum decon_dsi_mode dsi_mode, u32 width, u32 height)
{
	u32 val;

	if (dsi_mode == DSI_MODE_DUAL_DSI)
		width = width / 2;

	val = FRAME_FIFO_HEIGHT_F(height) | FRAME_FIFO_WIDTH_F(width);
	decon_write(id, FRAME_FIFO_0_SIZE_CONTROL_0, val);

	val = height * width;
	decon_write(id, FRAME_FIFO_0_SIZE_CONTROL_1, val);

	if (dsi_mode == DSI_MODE_DUAL_DSI) {
		val = FRAME_FIFO_HEIGHT_F(height) | FRAME_FIFO_WIDTH_F(width);
		decon_write(id, FRAME_FIFO_1_SIZE_CONTROL_0, val);

		val = height * width;
		decon_write(id, FRAME_FIFO_1_SIZE_CONTROL_1, val);
	}
}

void decon_reg_get_frame_fifo_size(u32 id, int dsi_idx, u32 *w, u32 *h)
{
	u32 val;

	if (id == 0) {
		if (dsi_idx == 0)
			val = decon_read(id, FRAME_FIFO_0_SIZE_CONTROL_0);
		else if (dsi_idx == 1)
			val = decon_read(id, FRAME_FIFO_1_SIZE_CONTROL_0);
	} else {
		/* TODO: Other DECON case will be implemented */
	}

	*w = FRAME_FIFO_WIDTH_GET(val);
	*h = FRAME_FIFO_HEIGHT_GET(val);
}

void decon_reg_set_splitter(u32 id, enum decon_dsi_mode dsi_mode, u32 width, u32 height)
{
	u32 val;
	u32 cr_0_setting_guide = 0;	/* 0 : guide about 1/2 mic, 1 : UM's guide about 1/2 mic */

	if (id == 0) {
		if (dsi_mode == DSI_MODE_DUAL_DSI) {

			if (cr_0_setting_guide == 0) {
				decon_write(id, SPLITTER_CONTROL_0, (width / 2));

				val = SPLITTER_HEIGHT_F(height) | SPLITTER_WIDTH_F(width);
				decon_write(id, SPLITTER_SIZE_CONTROL_0, val);

				val = height * width;
				decon_write(id, SPLITTER_SIZE_CONTROL_1, val);
			} else {
				decon_write(id, SPLITTER_CONTROL_0, width);

				val = SPLITTER_HEIGHT_F(height) | SPLITTER_WIDTH_F((width) * 2);
				decon_write(id, SPLITTER_SIZE_CONTROL_0, val);

				val = height * width * 2;
				decon_write(id, SPLITTER_SIZE_CONTROL_1, val);
			}
		} else {
			decon_write(id, SPLITTER_CONTROL_0, 0);

			val = SPLITTER_HEIGHT_F(height) | SPLITTER_WIDTH_F(width);
			decon_write(id, SPLITTER_SIZE_CONTROL_0, val);

			val = height * width;
			decon_write(id, SPLITTER_SIZE_CONTROL_1, val);
		}
	}
}

void decon_reg_get_splitter_size(u32 id, u32 *w, u32 *h)
{
	u32 val;

	val = decon_read(id, SPLITTER_SIZE_CONTROL_0);

	*w = SPLITTER_WIDTH_GET(val);
	*h = SPLITTER_HEIGHT_GET(val);
}

void decon_reg_set_dispif_porch(u32 id, int dsi_idx, struct decon_lcd *lcd_info)
{
	u32 val;

	val = DISPIF_TIMING_VBPD_F(lcd_info->vbp) | DISPIF_TIMING_VFPD_F(lcd_info->vfp);
	if (id == 0) {
		if (dsi_idx == 0)
			decon_write(0, DISPIF0_TIMING_CONTROL_0, val);
		else if (dsi_idx == 1)
			decon_write(0, DISPIF1_TIMING_CONTROL_0, val);
	}

	val = DISPIF_TIMING_VSPD_F(lcd_info->vsa);
	if (id == 0) {
		if (dsi_idx == 0)
			decon_write(0, DISPIF0_TIMING_CONTROL_1, val);
		else if (dsi_idx == 1)
			decon_write(0, DISPIF1_TIMING_CONTROL_1, val);
	}

	val = DISPIF_TIMING_HBPD_F(lcd_info->hbp) | DISPIF_TIMING_HFPD_F(lcd_info->hfp);
	if (id == 0) {
		if (dsi_idx == 0)
			decon_write(0, DISPIF0_TIMING_CONTROL_2, val);
		else if (dsi_idx == 1)
			decon_write(0, DISPIF1_TIMING_CONTROL_2, val);
	}

	val = DISPIF_TIMING_HSPD_F(lcd_info->hsa);
	if (id == 0) {
		if (dsi_idx == 0)
			decon_write(0, DISPIF0_TIMING_CONTROL_3, val);
		else if (dsi_idx == 1)
			decon_write(0, DISPIF1_TIMING_CONTROL_3, val);
	}
}

void decon_reg_set_dispif_size(u32 id, enum decon_dsi_mode dsi_mode, int dsi_idx, u32 width, u32 height)
{
	u32 val;

	if (dsi_mode == DSI_MODE_DUAL_DSI)
		width = width / 2;

	val = DISPIF_HEIGHT_F(height) | DISPIF_WIDTH_F(width);
	if (id == 0) {
		if (dsi_idx == 0)
			decon_write(0, DISPIF0_SIZE_CONTROL_0, val);
		else if (dsi_idx == 1)
			decon_write(0, DISPIF1_SIZE_CONTROL_0, val);
	}

	val = width * height;
	if (id == 0) {
		if (dsi_idx == 0)
			decon_write(0, DISPIF0_SIZE_CONTROL_1, val);
		else if (dsi_idx == 1)
			decon_write(0, DISPIF1_SIZE_CONTROL_1, val);
	}
}

void decon_reg_get_dispif_size(u32 id, int dsi_idx, u32 *p_width, u32 *p_height)
{
	u32 val;

	if (id == 0) {
		if (dsi_idx == 0)
			val = decon_read(0, DISPIF0_SIZE_CONTROL_0);
		else if (dsi_idx == 1)
			val = decon_read(0, DISPIF1_SIZE_CONTROL_0);
	}

	*p_width = DISPIF_WIDTH_GET(val);
	*p_height = DISPIF_HEIGHT_GET(val);
}

void decon_reg_update_req_global(u32 id)
{
	decon_write_mask(id, SHADOW_REG_UPDATE_REQ, ~0, SHADOW_REG_UPDATE_REQ_GLOBAL);
}

void decon_reg_update_req_window(u32 id, u32 win_idx)
{
	decon_write_mask(id, SHADOW_REG_UPDATE_REQ, ~0, SHADOW_REG_UPDATE_REQ_WIN(win_idx));
}

void decon_reg_all_win_shadow_update_req(u32 id)
{
	u32 mask;

	/* DECON_F has 8 windows */
	if (!id)
		mask = SHADOW_REG_UPDATE_REQ_FOR_DECON_F;

	decon_write_mask(id, SHADOW_REG_UPDATE_REQ, ~0, mask);
}

void decon_reg_direct_on_off(u32 id, u32 en)
{
	u32 val = en ? ~0 : 0;

	decon_write_mask(id, GLOBAL_CONTROL, val, GLOBAL_CONTROL_DECON_EN | GLOBAL_CONTROL_DECON_EN_F);
}

void decon_reg_per_frame_off(u32 id)
{
	decon_write_mask(id, GLOBAL_CONTROL, 0, GLOBAL_CONTROL_DECON_EN_F);
}

void decon_reg_configure_lcd(u32 id, enum decon_dsi_mode dsi_mode, struct decon_lcd *lcd_info)
{
	enum decon_data_path d_path;
	enum decon_enhance_path e_path;

	decon_reg_set_rgb_order(id, DECON_RGB);

	decon_reg_set_dispif_porch(id, 0, lcd_info);

	if (dsi_mode == DSI_MODE_DUAL_DSI)
		decon_reg_set_dispif_porch(id, 1, lcd_info);

	if (lcd_info->dsc_enabled) {
		if (lcd_info->dsc_cnt == 1)
			decon_reg_set_data_path(id,
					DATAPATH_ENC0_SPLITTERBYPASS_U0FF_DISP0,
					ENHANCEPATH_ENHANCE_ALL_OFF);
		else
			decon_err("not supported data path for %d DSC of decon%d",
					lcd_info->dsc_cnt, id);

		dsc_reg_init(id, dsi_mode, lcd_info);
	} else {
		decon_reg_set_splitter(id, dsi_mode, lcd_info->xres, lcd_info->yres);
		decon_reg_set_frame_fifo_size(id, dsi_mode, lcd_info->xres, lcd_info->yres);
		decon_reg_set_dispif_size(id, dsi_mode, 0, lcd_info->xres, lcd_info->yres);
		if (dsi_mode == DSI_MODE_DUAL_DSI)
			decon_reg_set_dispif_size(id, dsi_mode, 1, lcd_info->xres, lcd_info->yres);

		if (dsi_mode == DSI_MODE_DUAL_DSI)
			decon_reg_set_data_path(id, DATAPATH_NOCOMP_SPLITTER_U0FFU1FF_DISP0DISP1, ENHANCEPATH_ENHANCE_ALL_OFF);
		else
			decon_reg_set_data_path(id, DATAPATH_NOCOMP_SPLITTERBYPASS_U0FF_DISP0, ENHANCEPATH_ENHANCE_ALL_OFF);
	}

	decon_reg_set_vclk_freerun(id, 1);
	decon_reg_direct_on_off(id, 0);

	decon_reg_get_data_path(id, &d_path, &e_path);
	decon_dbg("%s: enhance path(0x%x), data path(0x%x)\n",
			__func__, e_path, d_path);
}

void decon_reg_configure_trigger(u32 id, enum decon_trig_mode mode)
{
	u32 val, mask;

	mask = HW_SW_TRIG_CONTROL_HW_TRIG_EN;

	if (mode == DECON_SW_TRIG) {
		/* To support per-frameoff TRIG_AUTO_MASK disabled */
		mask |= HW_SW_TRIG_CONTROL_TRIG_AUTO_MASK_TRIG;
		val = 0;
	} else {
		val = ~0;
	}

	decon_write_mask(id, HW_SW_TRIG_CONTROL, val, mask);
}


u32 decon_reg_get_linecnt(u32 id, int dispif_idx)
{
	u32 val;

	if (dispif_idx == 0)
		val = decon_read_mask(0, DISPIF_LINE_COUNT, DISPIF_LINE_COUNT_DISPIF0_MASK) >> DISPIF_LINE_COUNT_DISPIF0_SHIFT;
	else if (dispif_idx == 1)
		val = decon_read_mask(0, DISPIF_LINE_COUNT, DISPIF_LINE_COUNT_DISPIF1_MASK) >> DISPIF_LINE_COUNT_DISPIF1_SHIFT;

	return val;
}

/* timeout : usec */
int decon_reg_wait_linecnt_is_zero_timeout(u32 id, int dsi_idx, unsigned long timeout)
{
	unsigned long delay_time = 100;
	int disp_idx = (!id) ? id : (id + 1);
	unsigned long cnt = timeout / delay_time;
	u32 linecnt;

	do {
		linecnt = decon_reg_get_linecnt(id, disp_idx);
		if (!linecnt)
			break;
		cnt--;
		udelay(delay_time);
	} while (cnt);

	if (!cnt) {
		decon_err("wait timeout linecount is zero(%u)\n", linecnt);
		return -EBUSY;
	}

	return 0;
}

u32 decon_reg_get_idle_status(u32 id)
{
	u32 val;

	val = decon_read(id, GLOBAL_CONTROL);
	if (val & GLOBAL_CONTROL_IDLE_STATUS)
		return 1;

	return 0;
}

int decon_reg_wait_idle_status_timeout(u32 id, unsigned long timeout)
{
	unsigned long delay_time = 10;
	unsigned long cnt = timeout / delay_time;
	u32 status;

	do {
		status = decon_reg_get_idle_status(id);
		cnt--;
		udelay(delay_time);
	} while (!status && cnt);

	if (!cnt) {
		decon_err("wait timeout decon idle status(%u)\n", status);
		return -EBUSY;
	}

	return 0;
}

u32 decon_reg_get_run_status(u32 id)
{
	u32 val;

	val = decon_read(id, GLOBAL_CONTROL);
	if (val & GLOBAL_CONTROL_RUN_STATUS)
		return 1;

	return 0;
}

/* Determine that DECON is perfectly shuttled off through checking this * function */
int decon_reg_wait_run_is_off_timeout(u32 id, unsigned long timeout)
{
	unsigned long delay_time = 10;
	unsigned long cnt = timeout / delay_time;
	u32 status;

	do {
		status = decon_reg_get_run_status(id);
		cnt--;
		udelay(delay_time);
	} while (status && cnt);

	if (!cnt) {
		decon_err("wait timeout decon run is shut-off(%u)\n", status);
		return -EBUSY;
	}

	return 0;
}

int decon_reg_wait_run_status_timeout(u32 id, unsigned long timeout)
{
	unsigned long delay_time = 10;
	unsigned long cnt = timeout / delay_time;
	u32 status;

	do {
		status = decon_reg_get_run_status(id);
		cnt--;
		udelay(delay_time);
	} while (!status && cnt);

	if (!cnt) {
		decon_err("wait timeout decon run status(%u)\n", status);
		return -EBUSY;
	}

	return 0;
}

void decon_reg_clear_int_all(u32 id)
{
	u32 mask;

	/* mask = INTERRUPT_DISPIF_VSTATUS | INTERRUPT_RESOURCE_CONFLICT | INTERRUPT_FRAME_DONE | INTERRUPT_FIFO_LEVEL; */
	mask = ~0;
	decon_write_mask(id, INTERRUPT_PENDING, ~0, mask);
}

u32 decon_get_win_channel(u32 id, u32 win_idx, enum decon_idma_type type)
{
	u32 ch_id;

	switch (type) {
	case IDMA_G1:
		ch_id = 0;
		break;
	case IDMA_G2:
		ch_id = 1;
		break;
	case IDMA_G0:
		ch_id = 2;
		break;
	case IDMA_VG0:
		ch_id = 3;
		break;
	default:
		decon_dbg("channel(0x%x) is not valid\n", type);
		return -EINVAL;
	}

	return ch_id;
}

void decon_reg_config_win_channel(u32 id, u32 win_idx, enum decon_idma_type type)
{
	u32 ch_id;

	ch_id = decon_get_win_channel(id, win_idx, type);

	decon_write_mask(id, WIN_CONTROL(win_idx), WIN_CONTROL_CHMAP_F(ch_id), WIN_CONTROL_CHMAP_MASK);
}


/* wait until shadow update is finished */
int decon_reg_wait_for_update_timeout(u32 id, unsigned long timeout)
{
	unsigned long delay_time = 100;
	unsigned long cnt = timeout / delay_time;

	while (decon_read(id, SHADOW_REG_UPDATE_REQ) && --cnt)
		udelay(delay_time);

	if (!cnt) {
		decon_err("timeout of updating decon registers\n");
		return -EBUSY;
	}

	return 0;
}


/* wait until shadow update is finished */
int decon_reg_wait_for_window_update_timeout(u32 id, u32 win_idx, unsigned long timeout)
{
	unsigned long delay_time = 100;
	unsigned long cnt = timeout / delay_time;

	while ((decon_read(id, SHADOW_REG_UPDATE_REQ) & SHADOW_REG_UPDATE_REQ_WIN(win_idx)) && --cnt)
		udelay(delay_time);

	if (!cnt) {
		decon_err("timeout of updating decon window registers\n");
		return -EBUSY;
	}

	return 0;
}



/* Is it need to do hw trigger unmask and mask asynchronously in case of dual DSI */
/* enable(unmask) / disable(mask) hw trigger */
void decon_reg_set_trigger(u32 id, struct decon_mode_info *psr, enum decon_set_trig en)
{
	u32 val = (en == DECON_TRIG_ENABLE) ? ~0 : 0;
	u32 mask;

	if (psr->psr_mode == DECON_VIDEO_MODE)
		return;

	if (psr->trig_mode == DECON_SW_TRIG) {
		val = (en == DECON_TRIG_ENABLE) ? ~0 : 0;
		mask = HW_SW_TRIG_CONTROL_SW_TRIG;
	} else {
		val = (en == DECON_TRIG_DISABLE) ? ~0 : 0;
		mask = HW_SW_TRIG_CONTROL_HW_TRIG_MASK;
	}

	decon_write_mask(id, HW_SW_TRIG_CONTROL, val, mask);
}

/****************** DSC related functions ********************/
void dsc_reg_set_reset(u32 encoder_id)
{
	u32 reg;

	reg =  dsc_read(encoder_id, DSC_CONTROL0);
	reg |= DSC_CONTROL0_SW_RESET;
	dsc_write(encoder_id, DSC_CONTROL0, reg);
}
void dsc_reg_set_swap(u32 encoder_id, u32 uBit, u32 uByte, u32 uWord)
{
	u32 reg;

	reg =  dsc_read(encoder_id, DSC_CONTROL0);
	reg &= ~((DSC_CONTROL0_BIT_SWAP_MASK | DSC_CONTROL0_BYTE_SWAP_MASK | DSC_CONTROL0_WORD_SWAP_MASK));
	reg |= (DSC_CONTROL0_BIT_SWAP(uBit) | DSC_CONTROL0_BYTE_SWAP(uByte) | DSC_CONTROL0_WORD_SWAP(uWord));
	dsc_write(encoder_id, DSC_CONTROL0, reg);
}
void dsc_reg_set_flatness_det_th(u32 encoder_id, u32 uThreshold)
{
	u32 reg;

	reg =  dsc_read(encoder_id, DSC_CONTROL0);
	reg &= ~(DSC_CONTROL0_FLATNESS_DET_TH_F_MASK);
	reg |= DSC_CONTROL0_FLATNESS_DET_TH_F(uThreshold);
	dsc_write(encoder_id, DSC_CONTROL0, reg);
}

void dsc_reg_set_slice_mode_change(u32 encoder_id, u32 uEn)
{
	u32 reg;

	reg =  dsc_read(encoder_id, DSC_CONTROL0);
	reg &= ~(DSC_CONTROL0_SLICE_MODE_CH_F_MASK);
	reg |= DSC_CONTROL0_SLICE_MODE_CH_F(uEn);
	dsc_write(encoder_id, DSC_CONTROL0, reg);
}
void dsc_reg_set_encoder_bypass(u32 encoder_id, u32 uEn)
{
	u32 reg;

	reg =  dsc_read(encoder_id, DSC_CONTROL0);
	reg &= ~(DSC_CONTROL0_DSC_BYPASS_F_MASK);
	reg |= DSC_CONTROL0_DSC_BYPASS_F(uEn);
	dsc_write(encoder_id, DSC_CONTROL0, reg);
}
void dsc_reg_set_auto_clock_gating(u32 encoder_id, u32 uEn)
{
	u32 reg;

	reg =  dsc_read(encoder_id, DSC_CONTROL0);
	reg &= ~(DSC_CONTROL0_DSC_CG_EN_F_MASK);
	reg |= DSC_CONTROL0_DSC_CG_EN_F(uEn);
	dsc_write(encoder_id, DSC_CONTROL0, reg);
}

void dsc_reg_set_dual_slice_mode(u32 encoder_id, u32 uEn)
{
	u32 reg;

	reg =  dsc_read(encoder_id, DSC_CONTROL0);
	reg &= ~(DSC_CONTROL0_DUAL_SLICE_EN_F_MASK);
	reg |= DSC_CONTROL0_DUAL_SLICE_EN_F(uEn);
	dsc_write(encoder_id, DSC_CONTROL0, reg);
}
void dsc_reg_set_vertical_porch(u32 encoder_id, u32 uVFP, u32 uVSW, u32 uVBP)
{
	u32 reg;

	reg = (DSC_CONTROL1_DSC_V_FRONT_F(uVFP) | DSC_CONTROL1_DSC_V_SYNC_F(uVSW) | DSC_CONTROL1_DSC_V_BACK_F(uVBP));
	dsc_write(encoder_id, DSC_CONTROL1, reg);
}
void dsc_reg_set_horizontal_porch(u32 encoder_id, u32 uHFP, u32 uHSW, u32 uHBP)
{
	u32 reg;

	reg = (DSC_CONTROL2_DSC_H_FRONT_F(uHFP) | DSC_CONTROL2_DSC_H_SYNC_F(uHSW) | DSC_CONTROL2_DSC_H_BACK_F(uHBP));
	dsc_write(encoder_id, DSC_CONTROL2, reg);
}

void dsc_reg_set_input_pixel_count(u32 encoder_id, u32 picture_width, u32 picture_height)
{
	u32 reg;

	reg = (picture_width*picture_height);
	dsc_write(encoder_id, DSC_IN_PIXEL_COUNT, reg);
}
void dsc_reg_set_comp_pixel_count(u32 encoder_id)
{
	u32 reg;
	reg = decon_read(0, FRAME_FIFO_0_SIZE_CONTROL_1);
	dsc_write(encoder_id, DSC_COMP_PIXEL_COUNT, reg);
}

/*
#########################################################################################
						dsc PPS Configuration
#########################################################################################
*/
u32 dsc_reg_get_num_extra_mux_bits(u32 SliceHeight, u32 chunk_size)
{
	u32 uNum_Extra_Mux_Bits;
	u32 uSliceBits;

	uNum_Extra_Mux_Bits = DSC_NUM_EXTRA_MUX_BIT;
	uSliceBits = 8*chunk_size*SliceHeight;
	while ((uSliceBits-uNum_Extra_Mux_Bits)%48)
		uNum_Extra_Mux_Bits--;

	return	uNum_Extra_Mux_Bits;
}
u32 dsc_reg_get_slice_height(u32 picture_height, u32 slice_width)
{

	u32 slice_height;
	u32 uMinSliceSize;
	u32 uQuotient = 1;
	u32 i, uCnt = 0;
	u32 uTemp[] = {0,};
	decon_dbg("\npicture_height:%d", picture_height);
	decon_dbg("\nslice_width:%d", slice_width);

	while (1) {
		if ((picture_height%uQuotient) == 0) {
			uTemp[uCnt] = uQuotient;
			uCnt++;
		}
		uQuotient++;
		if (picture_height/2 < uQuotient) {
			uTemp[uCnt-1] = picture_height;
			break;
		}
	}
	for (i = 0; i < uCnt; i++)
		decon_dbg("\nQuotient[%d]:%d", i, uTemp[i]);
	for (i = 0; i < uCnt; i++) {
		slice_height = uTemp[i];
		uMinSliceSize = slice_width*slice_height;
		if (uMinSliceSize > DSC_MIN_SLICE_SIZE)
			break;
	}

	decon_dbg("\n SliceHeight: %d", slice_height);
	return	slice_height;
}

void dsc_reg_set_pps_0_3_dsc_version(u32 encoder_id, u32 ver)
{
	u32 reg;

	reg = dsc_read(encoder_id, DSC_PPS00_03);
	reg &= ~DSC_PPS00_03_DSC_VER_MASK;
	reg |= DSC_PPS00_03_DSC_VER(ver);
	dsc_write(encoder_id, DSC_PPS00_03, reg);
}

void dsc_reg_set_pps_6_7_picture_height(u32 encoder_id, u32 height)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, DSC_PPS04_07);
	reg &= ~DSC_PPS06_07_PICTURE_HEIGHT_MASK;
	reg |= DSC_PPS06_07_PICTURE_HEIGHT(height);
	dsc_write(encoder_id, DSC_PPS04_07, reg);
}
void dsc_reg_set_pps_8_9_picture_width(u32 encoder_id, u32 width)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, DSC_PPS08_11);
	reg &= ~DSC_PPS08_9_PICTURE_WIDHT_MASK;
	reg |= DSC_PPS08_9_PICTURE_WIDHT(width);
	dsc_write(encoder_id, DSC_PPS08_11, reg);
}
void dsc_reg_set_pps_10_11_slice_height(u32 encoder_id, u32 height)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, DSC_PPS08_11);
	reg &= ~(DSC_PPS10_11_SLICE_HEIGHT_MASK);
	reg |= DSC_PPS10_11_SLICE_HEIGHT(height);
	dsc_write(encoder_id, DSC_PPS08_11, reg);
}
void dsc_reg_set_pps_12_13_slice_width(u32 encoder_id, u32 width)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, DSC_PPS12_15);
	reg &= ~DSC_PPS12_13_SLICE_WIDTH_MASK;
	reg |= DSC_PPS12_13_SLICE_WIDTH(width);
	dsc_write(encoder_id, DSC_PPS12_15, reg);
}
void dsc_reg_set_pps_14_15_chunk_size(u32 encoder_id, u32 slice_width)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, DSC_PPS12_15);
	reg &= ~DSC_PPS14_15_CHUNK_SIZE_MASK;
	reg |= DSC_PPS14_15_CHUNK_SIZE(slice_width);
	dsc_write(encoder_id, DSC_PPS12_15, reg);
}
void dsc_reg_set_pps_16_17_initial_xmit_delay(u32 encoder_id)
{
	u32 reg;
	u32 uPPS16_InitialXmitDelay = 0x200;

	reg  =  dsc_read(encoder_id, DSC_PPS16_19);
	reg &= ~(0x3ff<<16);
	reg |= (uPPS16_InitialXmitDelay<<16);

	reg = 0x020001B4;
	dsc_write(encoder_id, DSC_PPS16_19, reg);
}
void dsc_reg_set_pps_18_19_initial_dec_delay(u32 encoder_id)
{
	//decon_info("\nPPS18, PPS19 Fixed value : PPS18(0x4), PPS19(0xC0)");
}
void dsc_reg_set_pps_21_initial_scale_value(u32 encoder_id)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, DSC_PPS20_23);
	reg &= ~DSC_PPS21_INIT_SCALE_VALUE_MASK;
	reg |= DSC_PPS21_INIT_SCALE_VALUE(DSC_INIT_SCALE_VALUE);
	dsc_write(encoder_id, DSC_PPS20_23, reg);
}

void dsc_reg_set_pps_22_23_scale_increment_interval(u32 encoder_id, u32 slice_width, u32 slice_height)
{
	u32 reg;
	u32 uPPS22_ScaleIncrementInterval;
	u32 uPPS28_NflBpgOffset;
	u32 uPPS30_SliceBpgOffset;
	u32 uPPS34_FinalOffset;
	u32 uFinalScale;
	u32 uGroupsTotal;
	u32 uNumExtraMuxBits;
	u32 chunk_size;
	u32 val;

#if 1
	val = (DSC_FIRST_LINE_BPG_OFFSET*2048) / (slice_height-1);
	uPPS28_NflBpgOffset = (u32)CEIL(val);
#else
	uPPS28_NflBpgOffset = (u32)CEIL((double)((DSC_FIRST_LINE_BPG_OFFSET*2048)/(slice_height-1)));
#endif
	uPPS28_NflBpgOffset += 1;

	chunk_size = slice_width;
	uNumExtraMuxBits = dsc_reg_get_num_extra_mux_bits(slice_height, chunk_size);
	uGroupsTotal = ((slice_width+2)/3)*slice_height;
	uPPS30_SliceBpgOffset = (u32)CEIL(2048*(DSC_RC_MODE_SIZE-DSC_INIT_OFFSET+uNumExtraMuxBits)/uGroupsTotal);

	uNumExtraMuxBits = dsc_reg_get_num_extra_mux_bits(slice_height, chunk_size);
	uPPS34_FinalOffset = DSC_RC_MODE_SIZE-((DSC_INIT_TRANSMIT_DELAY*DSC_BIT_PER_PIXEL+8)>>4)+uNumExtraMuxBits;

	uFinalScale	= 8*(DSC_RC_MODE_SIZE/(DSC_RC_MODE_SIZE-uPPS34_FinalOffset));

	uPPS22_ScaleIncrementInterval = (u32)(2048*uPPS34_FinalOffset/((uFinalScale-9)*(uPPS28_NflBpgOffset+uPPS30_SliceBpgOffset)));
	//decon_info("\nuPPS22_ScaleIncrementInterval=2048x(%d)/((%d-9x)(%d+%d))=(%d)", uPPS34_FinalOffset, uFinalScale, uPPS28_NflBpgOffset, uPPS30_SliceBpgOffset, uPPS22_ScaleIncrementInterval);

	reg  =  dsc_read(encoder_id, DSC_PPS20_23);
	reg &= ~DSC_PPS22_23_SCALE_INCREMENT_INTERVAL_MASK;
	reg |= DSC_PPS22_23_SCALE_INCREMENT_INTERVAL(0x04F2);
	dsc_write(encoder_id, DSC_PPS20_23, reg);
}

void dsc_reg_set_pps_24_25_scale_decrement_interval(u32 encoder_id, u32 slice_width)
{
	u32 reg;
	u32 uGroupPerLine;
	u32 uPPS24_ScaleDecrementInterval;

	uGroupPerLine = ((slice_width+2)/3);
	uPPS24_ScaleDecrementInterval = (uGroupPerLine/(DSC_INIT_SCALE_VALUE-8));

	reg  =  dsc_read(encoder_id, DSC_PPS24_27);
	reg &= ~DSC_PPS24_25_SCALE_DECREMENT_INTERVAL_MASK;
	reg |= DSC_PPS24_25_SCALE_DECREMENT_INTERVAL(uPPS24_ScaleDecrementInterval);
	dsc_write(encoder_id, DSC_PPS24_27, reg);
}

void dsc_reg_set_pps_27_first_line_bpg_offset(u32 encoder_id)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, DSC_PPS24_27);
	reg &= ~DSC_PPS27_FIRST_LINE_BPG_OFFSET_MASK;
	reg |= DSC_PPS27_FIRST_LINE_BPG_OFFSET(DSC_FIRST_LINE_BPG_OFFSET);
	dsc_write(encoder_id, DSC_PPS24_27, reg);
}

void dsc_reg_set_pps_28_29_nfl_bpg_offset(u32 encoder_id, u32 slice_height)
{
	u32 reg;
	u32 uPPS28_uNflBpgOffset;

	uPPS28_uNflBpgOffset = (u32)CEIL(((DSC_FIRST_LINE_BPG_OFFSET*2048)/(slice_height-1)));
	uPPS28_uNflBpgOffset += 1;

	reg  =  dsc_read(encoder_id, DSC_PPS28_31);
	reg &= ~DSC_PPS28_29_NOT_FIRST_LINE_BPG_OFFSET_MASK;
	reg |= DSC_PPS28_29_NOT_FIRST_LINE_BPG_OFFSET(uPPS28_uNflBpgOffset);
	dsc_write(encoder_id, DSC_PPS28_31, reg);
}

void dsc_reg_set_pps_30_31_slice_bpg_offset(u32 encoder_id, u32 slice_width, u32 slice_height, u32 chunk_size)
{
	u32 reg;
	u32 uPPS30_SliceBpgOffset;
	u32 uNumExtraMuxBits;
	u32 uGroupsTotal;

	uNumExtraMuxBits = dsc_reg_get_num_extra_mux_bits(slice_height, chunk_size);
	uGroupsTotal = ((slice_width+2)/3)*slice_height;
	uPPS30_SliceBpgOffset = (u32)CEIL(2048*(DSC_RC_MODE_SIZE-DSC_INIT_OFFSET+uNumExtraMuxBits)/uGroupsTotal);

	reg  =  dsc_read(encoder_id, (DSC_PPS28_31));
	reg &= ~DSC_PPS30_31_SLICE_BPG_OFFSET_MASK;
	reg |= DSC_PPS30_31_SLICE_BPG_OFFSET(0x0263);
	dsc_write(encoder_id, (DSC_PPS28_31), reg);
}

void dsc_reg_set_pps_32_33_initial_offset(u32 encoder_id)
{
	u32 reg;

	reg  =  dsc_read(encoder_id, (DSC_PPS32_35));
	reg &= ~DSC_PPS32_33_INIT_OFFSET_MASK;
	reg |= DSC_PPS32_33_INIT_OFFSET(DSC_INIT_OFFSET);
	dsc_write(encoder_id, (DSC_PPS32_35), reg);
}

void dsc_reg_set_pps_34_35_final_offset(u32 encoder_id, u32 chunk_size, u32 slice_height)
{
	u32 reg;
	u32 uPPS34_FinalOffset;
	u32 uNumExtraMuxBits;

	uNumExtraMuxBits = dsc_reg_get_num_extra_mux_bits(slice_height, chunk_size);
	uPPS34_FinalOffset = DSC_RC_MODE_SIZE-((DSC_INIT_TRANSMIT_DELAY*DSC_BIT_PER_PIXEL+8)>>4)+uNumExtraMuxBits;

	reg  =  dsc_read(encoder_id, (DSC_PPS32_35));
	reg &= ~DSC_PPS34_35_FINAL_OFFSET_MASK;
	reg |= DSC_PPS34_35_FINAL_OFFSET(uPPS34_FinalOffset);
	dsc_write(encoder_id, (DSC_PPS32_35), reg);
}

void dsc_reg_set_pps(u32 encoder_id, struct decon_lcd *lcd_info)
{
	u32 picture_w, picture_h;
	u32 slice_w, slice_h;
	u32 chunk_size;
	u32 dual_slice = 0;

	picture_h = lcd_info->yres;
	slice_w = lcd_info->xres / lcd_info->dsc_slice_num;
	slice_h = (lcd_info->dsc_slice_num == 4) ? 64 : 16;
	chunk_size = slice_w;

	if (lcd_info->dsc_cnt == 1) {
		picture_w = lcd_info->xres;
	} else if (lcd_info->dsc_cnt == 2) {
		picture_w = lcd_info->xres >> 1;
	} else {
		picture_w = lcd_info->xres;
		decon_err("DSC max count is %d\n", lcd_info->dsc_cnt);
	}

	if (lcd_info->dsc_slice_num == 4 ||
			(lcd_info->dsc_cnt == 1 && lcd_info->dsc_slice_num == 2))
		dual_slice = 1;

	decon_dbg("DSC%d: picture w(%d) h(%d)\n", encoder_id, picture_w, picture_h);
	decon_dbg("slice w(%d) h(%d), chunk(%d), dual slice(%d)\n",
			slice_w, slice_h, chunk_size, dual_slice);

	dsc_reg_set_swap(encoder_id, 0x0, 0x0, 0x0);
	dsc_reg_set_flatness_det_th(encoder_id, 0x2);
	dsc_reg_set_auto_clock_gating(encoder_id, 1);
	dsc_reg_set_dual_slice_mode(encoder_id, dual_slice);
	dsc_reg_set_encoder_bypass(encoder_id, 0);
	dsc_reg_set_slice_mode_change(encoder_id, 0);
	dsc_reg_set_input_pixel_count(encoder_id, picture_w, picture_h);
	dsc_reg_set_comp_pixel_count(encoder_id);

	dsc_reg_set_pps_0_3_dsc_version(encoder_id, 0x11);
	dsc_reg_set_pps_6_7_picture_height(encoder_id, picture_h);
	dsc_reg_set_pps_8_9_picture_width(encoder_id, picture_w);
	dsc_reg_set_pps_10_11_slice_height(encoder_id, slice_h);
	dsc_reg_set_pps_12_13_slice_width(encoder_id, slice_w);
	dsc_reg_set_pps_14_15_chunk_size(encoder_id, slice_w);
	dsc_reg_set_pps_16_17_initial_xmit_delay(encoder_id);
	dsc_reg_set_pps_18_19_initial_dec_delay(encoder_id);
	dsc_reg_set_pps_21_initial_scale_value(encoder_id);
	dsc_reg_set_pps_22_23_scale_increment_interval(encoder_id, slice_w, slice_h);
	dsc_reg_set_pps_24_25_scale_decrement_interval(encoder_id, slice_w);
	dsc_reg_set_pps_27_first_line_bpg_offset(encoder_id);
	dsc_reg_set_pps_28_29_nfl_bpg_offset(encoder_id, slice_h);
	dsc_reg_set_pps_30_31_slice_bpg_offset(encoder_id, slice_w, slice_h, chunk_size);
	dsc_reg_set_pps_32_33_initial_offset(encoder_id);
	dsc_reg_set_pps_34_35_final_offset(encoder_id, chunk_size, slice_h);
}

int dsc_reg_init(u32 id, enum decon_dsi_mode dsi_mode, struct decon_lcd *lcd_info)
{
	u32 width = lcd_info->xres / 3;
	u32 fifo_w = width;
	u32 w, h;

	if (lcd_info->dsc_cnt == 2) {
		fifo_w = width / 2;
		decon_reg_set_frame_fifo_size(id, DSI_MODE_DUAL_DSI, fifo_w,
				lcd_info->yres);
		dsc_reg_set_pps(1, lcd_info);
	}
	decon_reg_set_frame_fifo_size(id, DSI_MODE_SINGLE, fifo_w, lcd_info->yres);

	decon_reg_set_splitter(id, DSI_MODE_SINGLE, width, lcd_info->yres);
	decon_reg_set_dispif_size(id, DSI_MODE_SINGLE, 0, width, lcd_info->yres);
	dsc_reg_set_pps(0, lcd_info);

	decon_reg_get_splitter_size(id, &w, &h);
	decon_dbg("SPLITTER size: w(%d), h(%d)\n", w, h);
	decon_reg_get_frame_fifo_size(id, 0, &w, &h);
	decon_dbg("FIFO0 size: w(%d), h(%d)\n", w, h);
	decon_reg_get_frame_fifo_size(id, 1, &w, &h);
	decon_dbg("FIFO1 size: w(%d), h(%d)\n", w, h);
	decon_reg_get_dispif_size(id, 0, &w, &h);
	decon_dbg("DISPIF0 size: w(%d), h(%d)\n", w, h);

	return 0;
}

/***************** CAL APIs implementation *******************/
int decon_reg_init(u32 id, u32 dsi_idx, struct decon_param *p)
{
	struct decon_lcd *lcd_info = p->lcd_info;
	struct decon_mode_info *psr = &p->psr;

	/* DECON does not need to start, if DECON is already
	 * running(enabled in LCD_ON_UBOOT) */
#if 0
	if (decon_reg_get_run_status(id)) {
		decon_info("decon_reg_init already called by BOOTLOADER\n");
		decon_reg_init_probe(id, dsi_idx, p);
		if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
			decon_reg_set_trigger(id, psr, DECON_TRIG_DISABLE);
		return -EBUSY;
	}
#endif

	decon_reg_reset(id);

	decon_reg_set_clkgate_mode(id, 1);

	if (psr->dsi_mode == DSI_MODE_DUAL_DSI)
		/* (denom, num) : VCLK=VCLK_SRC * (num+1)/(denom+1) */
		decon_reg_set_vclk_divider(id, 2, 1);
	else
		/* (denom, num) : VCLK=VCLK_SRC * (num+1)/(denom+1) */
		decon_reg_set_vclk_divider(id, 1, 1);

	decon_reg_set_operation_mode(id, psr->psr_mode);

	decon_reg_set_blender_bg_image_size(id, psr->dsi_mode, lcd_info);

	if (id == 2) {
		/* Set a TRIG mode */
		decon_reg_configure_trigger(id, psr->trig_mode);

		/*
		 * Interrupt of DECON-T should be set to video mode,
		 * because of malfunction of I80 frame done interrupt.
		 */
		psr->psr_mode = DECON_VIDEO_MODE;
		decon_reg_set_int(id, psr, 1);
		decon_reg_set_underrun_scheme(id, DECON_VCLK_NOT_AFFECTED);

		/* RGB order -> frame fifo size -> splitter -> porch values -> mic configuration -> freerun mode --> stop DECON */
		decon_reg_configure_lcd(id, psr->dsi_mode, lcd_info);
	} else {
		if (psr->psr_mode == DECON_MIPI_COMMAND_MODE) {
			/* guided by designer to change from DECON_VCLK_RUNNING_VDEN_DISABLE */
			decon_reg_set_underrun_scheme(id, DECON_VCLK_HOLD_ONLY);
		} else {
			decon_reg_set_underrun_scheme(id, DECON_VCLK_HOLD_ONLY);
		}
		/* RGB order -> frame fifo size -> splitter -> porch values -> mic configuration -> freerun mode --> stop DECON */
		decon_reg_configure_lcd(id, psr->dsi_mode, lcd_info);

		if (psr->psr_mode == DECON_MIPI_COMMAND_MODE) {
			decon_reg_configure_trigger(id, psr->trig_mode);
			decon_reg_set_trigger(id, psr, DECON_TRIG_DISABLE);
		}
	}

	/* asserted interrupt should be cleared before initializing decon hw */
	decon_reg_clear_int_all(id);

	return 0;
}


void decon_reg_init_probe(u32 id, u32 dsi_idx, struct decon_param *p)
{
	struct decon_lcd *lcd_info = p->lcd_info;
	struct decon_mode_info *psr = &p->psr;
	enum decon_data_path d_path;
	enum decon_enhance_path e_path;

	decon_reg_set_clkgate_mode(id, 1);

	if (psr->dsi_mode == DSI_MODE_DUAL_DSI)
		decon_reg_set_vclk_divider(id, 2, 1);	/* (denom, num) : VCLK=VCLK_SRC * (num+1)/(denom+1) */
	else
		decon_reg_set_vclk_divider(id, 1, 1);	/* (denom, num) : VCLK=VCLK_SRC * (num+1)/(denom+1) */

	decon_reg_set_operation_mode(id, psr->psr_mode);

	decon_reg_set_blender_bg_image_size(id, psr->dsi_mode, lcd_info);

	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE) {
		/* guided by designer to change from DECON_VCLK_RUNNING_VDEN_DISABLE */
		decon_reg_set_underrun_scheme(id, DECON_VCLK_HOLD_ONLY);
	} else {
		decon_reg_set_underrun_scheme(id, DECON_VCLK_HOLD_ONLY);
	}
	/*
	* ->
	* same as decon_reg_configure_lcd(...) function
	* except using decon_reg_update_req_global(id) instead of decon_reg_direct_on_off(id, 0)
	*/
	decon_reg_set_rgb_order(id, DECON_RGB);

	decon_reg_set_dispif_porch(id, 0, lcd_info);

	if (psr->dsi_mode == DSI_MODE_DUAL_DSI)
		decon_reg_set_dispif_porch(id, 1, lcd_info);

	if (lcd_info->dsc_enabled) {
		if (lcd_info->dsc_cnt == 1)
			decon_reg_set_data_path(id,
					DATAPATH_ENC0_SPLITTERBYPASS_U0FF_DISP0,
					ENHANCEPATH_ENHANCE_ALL_OFF);
		else
			decon_err("not supported data path for %d DSC of decon%d",
					lcd_info->dsc_cnt, id);

		dsc_reg_init(id, psr->dsi_mode, lcd_info);
	} else {
		decon_reg_set_splitter(id, psr->dsi_mode, lcd_info->xres, lcd_info->yres);
		decon_reg_set_frame_fifo_size(id, psr->dsi_mode, lcd_info->xres, lcd_info->yres);
		decon_reg_set_dispif_size(id, psr->dsi_mode, 0, lcd_info->xres, lcd_info->yres);
		if (psr->dsi_mode == DSI_MODE_DUAL_DSI)
			decon_reg_set_dispif_size(id, psr->dsi_mode, 1, lcd_info->xres, lcd_info->yres);

		if (psr->dsi_mode == DSI_MODE_DUAL_DSI)
			decon_reg_set_data_path(id, DATAPATH_NOCOMP_SPLITTER_U0FFU1FF_DISP0DISP1, ENHANCEPATH_ENHANCE_ALL_OFF);
		else
			decon_reg_set_data_path(id, DATAPATH_NOCOMP_SPLITTERBYPASS_U0FF_DISP0, ENHANCEPATH_ENHANCE_ALL_OFF);
	}

	decon_reg_set_vclk_freerun(id, 1);
	decon_reg_update_req_global(id);
	/*
	* <-
	*/

	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE) {
		decon_reg_configure_trigger(id, psr->trig_mode);
		decon_reg_set_trigger(id, psr, DECON_TRIG_DISABLE);
	}

	decon_reg_get_data_path(id, &d_path, &e_path);
	decon_dbg("%s: enhance path(0x%x), data path(0x%x)\n",
			__func__, e_path, d_path);
}


int decon_reg_start(u32 id, struct decon_mode_info *psr)
{
	int ret = 0;
	decon_reg_direct_on_off(id, 1);

	decon_reg_update_req_global(id);

	/* DECON goes to run-status as soon as request shadow update without HW_TE */
	ret = decon_reg_wait_run_status_timeout(id, 20 * 1000);

	/* wait until run-status, then trigger */
	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
		decon_reg_set_trigger(id, psr, DECON_TRIG_ENABLE);
	return ret;
}

void decon_reg_set_partial_update(u32 id, enum decon_dsi_mode dsi_mode,
		struct decon_lcd *lcd_info)
{
	decon_reg_set_blender_bg_image_size(id, dsi_mode, lcd_info);
	decon_reg_set_dispif_porch(id, 0, lcd_info);

	if (dsi_mode == DSI_MODE_DUAL_DSI)
		decon_reg_set_dispif_porch(id, 1, lcd_info);

	if (!lcd_info->mic_enabled) {
		decon_reg_set_splitter(id, dsi_mode, lcd_info->xres, lcd_info->yres);
		decon_reg_set_frame_fifo_size(id, dsi_mode, lcd_info->xres, lcd_info->yres);
		decon_reg_set_dispif_size(id, dsi_mode, 0, lcd_info->xres, lcd_info->yres);
		if (dsi_mode == DSI_MODE_DUAL_DSI)
			decon_reg_set_dispif_size(id, dsi_mode,  1, lcd_info->xres, lcd_info->yres);
	}
}

int decon_reg_stop(u32 id, u32 dsi_idx, struct decon_mode_info *psr)
{
	int ret = 0;

#if 0		/* method which I recommends */
	decon_reg_per_frame_off(id);
	decon_reg_update_req_global(id);

	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE) {
		if (psr->trig_mode == DECON_SW_TRIG) {
			/* s/w trigger will work in case vertical status is not active in codition TRIG_AUTO_MASK_EN is enabled */
			ret = decon_reg_wait_idle_status_timeout(id, 20 * 1000);	/* timeout : 20ms */
			if (ret)
				goto err;
		}
		decon_reg_set_trigger(id, psr, DECON_TRIG_ENABLE);
	}

	/* timeout : 20ms */
	ret = decon_reg_wait_run_is_off_timeout(id, 20 * 1000);
	if (ret)
		goto err;

	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
		decon_reg_set_trigger(id, psr, DECON_TRIG_DISABLE);

	decon_reg_per_frame_off(id);	/* UM guide : clear ENVID_F again in case of per-frame stop (because of Q-channel??) */

#else		/* method which SWSOL wants */
	if ((psr->psr_mode == DECON_MIPI_COMMAND_MODE) &&
			(psr->trig_mode == DECON_HW_TRIG)) {
		decon_reg_set_trigger(id, psr, DECON_TRIG_DISABLE);
	}

	/* timeout : 50ms */
	/* TODO: dual DSI scenario */
	ret = decon_reg_wait_linecnt_is_zero_timeout(id, dsi_idx, 50 * 1000);
	if (ret)
		goto err;

	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
		decon_reg_direct_on_off(id, 0);
	else
		decon_reg_per_frame_off(id);

	decon_reg_update_req_global(id);

	/* timeout : 20ms */
	ret = decon_reg_wait_run_is_off_timeout(id, 20 * 1000);
	if (ret)
		goto err;
#endif

err:
	ret = decon_reg_reset(id);

	return ret;
}

void decon_reg_release_resource(u32 id, struct decon_mode_info *psr)
{
	decon_reg_per_frame_off(id);
	decon_reg_update_req_global(id);
	decon_reg_set_trigger(id, psr, DECON_TRIG_ENABLE);
}

void decon_reg_set_int(u32 id, struct decon_mode_info *psr, u32 en)
{
	u32 val;

	decon_reg_clear_int_all(id);

#if 0
	if (en) {
		val = INTERRUPT_INT_EN | INTERRUPT_FIFO_LEVEL_INT_EN | INTERRUPT_RESOURCE_CONFLICT_INT_EN | INTERRUPT_DPU0_INT_EN | INTERRUPT_DPU1_INT_EN;
		if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
			val |= INTERRUPT_FRAME_DONE_INT_EN;
		else
			val |= INTERRUPT_DISPIF_VSTATUS_INT_EN | INTERRUPT_DISPIF_VSTATUS_VSA;
		decon_write_mask(id, INTERRUPT_ENABLE, val, ~0);
	} else {
		decon_write_mask(id, INTERRUPT_ENABLE, 0, INTERRUPT_INT_EN);
	}

#else	/* OS always use Vstatus Interrupt irrespective of operating mode */
	if (en) {
		val = INTERRUPT_INT_EN | INTERRUPT_FIFO_LEVEL_INT_EN | INTERRUPT_DPU0_INT_EN;
		if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
			val |= INTERRUPT_FRAME_DONE_INT_EN;
			val |= INTERRUPT_DISPIF_VSTATUS_INT_EN | INTERRUPT_DISPIF_VSTATUS_VSA;
		decon_write_mask(id, INTERRUPT_ENABLE, val, ~0);
	} else {
		decon_write_mask(id, INTERRUPT_ENABLE, 0, INTERRUPT_INT_EN);
	}
#endif
}

void decon_reg_set_winmap(u32 id, u32 idx, u32 color, u32 en)
{
	u32 val = en ? WIN_CONTROL_MAPCOLOR_EN_F : 0;

	/* Enable */
	decon_write_mask(id, WIN_CONTROL(idx), val, WIN_CONTROL_MAPCOLOR_EN_MASK);

	/* Color Set */
	val = WIN_COLORMAP_MAPCOLOR_F(color);
	decon_write_mask(id, WIN_COLORMAP(idx), val, WIN_COLORMAP_MAPCOLOR_MASK);
}

void decon_reg_set_window_control(u32 id, int win_idx,
		struct decon_window_regs *regs, u32 winmap_en)
{
	decon_write(id, WIN_CONTROL(win_idx), regs->wincon);
	decon_write(id, WIN_START_POSITION(win_idx), regs->start_pos);
	decon_write(id, WIN_END_POSITION(win_idx), regs->end_pos);
	decon_write(id, WIN_START_TIME_CONTROL(win_idx), regs->start_time);
	decon_write(id, WIN_PIXEL_COUNT(win_idx), regs->pixel_count);

	decon_reg_set_winmap(id, win_idx, regs->colormap, winmap_en);
	decon_reg_config_win_channel(id, win_idx, regs->type);

	decon_dbg("%s: regs->type(%d)\n", __func__, regs->type);
}


void decon_reg_update_req_and_unmask(u32 id, struct decon_mode_info *psr)
{
	decon_reg_update_req_global(id);

	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
		decon_reg_set_trigger(id, psr, DECON_TRIG_ENABLE);
}


int decon_reg_wait_update_done_and_mask(u32 id, struct decon_mode_info *psr, u32 timeout)
{
	int result;

	result = decon_reg_wait_for_update_timeout(id, timeout);

	if (psr->psr_mode == DECON_MIPI_COMMAND_MODE)
		decon_reg_set_trigger(id, psr, DECON_TRIG_DISABLE);

	return result;
}

int decon_reg_get_interrupt_and_clear(u32 id)
{
	u32 val = decon_read(id, INTERRUPT_PENDING);

	if (val & INTERRUPT_FIFO_LEVEL_INT_EN)
		decon_write(id, INTERRUPT_PENDING, INTERRUPT_FIFO_LEVEL_INT_EN);

	if (val & INTERRUPT_FRAME_DONE_INT_EN)
		decon_write(id, INTERRUPT_PENDING, INTERRUPT_FRAME_DONE_INT_EN);

	if (val & INTERRUPT_DISPIF_VSTATUS_INT_EN)
		decon_write(id, INTERRUPT_PENDING, INTERRUPT_DISPIF_VSTATUS_INT_EN);

	if (val & INTERRUPT_DPU0_INT_EN)
		decon_write(id, INTERRUPT_PENDING, INTERRUPT_DPU0_INT_EN);

	return val;
}

/******************  OS Only ******************/
int decon_reg_is_win_enabled(u32 id, int win_idx)
{
	if (decon_read(id, WIN_CONTROL(win_idx)) & WIN_CONTROL_EN_F)
		return 1;

	return 0;
}

u32 decon_reg_get_width(u32 id, int dsi_mode)
{
	int disp_idx = (!id) ? id : (id + 1);
	u32 val = 0;

	val = decon_read(id, DISPIF_SIZE_CONTROL_0(disp_idx));
	return (val >> DISPIF_WIDTH_START_POS) & DISPIF_WIDTH_MASK;
}

u32 decon_reg_get_height(u32 id, int dsi_mode)
{
	int disp_idx = (!id) ? id : (id + 1);
	u32 val = 0;

	val = decon_read(id, DISPIF_SIZE_CONTROL_0(disp_idx));
	return (val >> DISPIF_HEIGHT_START_POS) & DISPIF_WIDTH_MASK;
}

const unsigned long decon_clocks_table[][CLK_ID_MAX] = {
	/* VCLK,  ECLK,  ACLK,  PCLK,  DISP_PLL,  resolution,           MIC_ratio, DSC count */
	{   266, 266,   400,    66,       266, 1440 * 2560,     MIC_COMP_BYPASS,         0},
};

void decon_reg_get_clock_ratio(struct decon_clocks *clks, struct decon_lcd *lcd_info)
{
	int i = sizeof(decon_clocks_table) / sizeof(decon_clocks_table[0]) - 1;

	/* set reset value */
	clks->decon[CLK_ID_VCLK] = decon_clocks_table[0][CLK_ID_VCLK];
	clks->decon[CLK_ID_ECLK] = decon_clocks_table[0][CLK_ID_ECLK];
	clks->decon[CLK_ID_ACLK] = decon_clocks_table[0][CLK_ID_ACLK];
	clks->decon[CLK_ID_PCLK] = decon_clocks_table[0][CLK_ID_PCLK];
	clks->decon[CLK_ID_DPLL] = decon_clocks_table[0][CLK_ID_DPLL];

	for (; i >= 0; i--) {
		if (decon_clocks_table[i][CLK_ID_RESOLUTION]
				!= lcd_info->xres * lcd_info->yres) {
			continue;
		}

		if (!lcd_info->mic_enabled && !lcd_info->dsc_enabled) {
			if (decon_clocks_table[i][CLK_ID_MIC_RATIO]
					!= MIC_COMP_BYPASS)
				continue;
		}

		if (lcd_info->mic_enabled) {
			if (decon_clocks_table[i][CLK_ID_MIC_RATIO]
					!= lcd_info->mic_ratio)
				continue;
		}

		if (lcd_info->dsc_enabled) {
			if (decon_clocks_table[i][CLK_ID_DSC_RATIO]
					!= lcd_info->dsc_cnt)
				continue;
		}

		clks->decon[CLK_ID_VCLK] = decon_clocks_table[i][CLK_ID_VCLK];
		clks->decon[CLK_ID_ECLK] = decon_clocks_table[i][CLK_ID_ECLK];
		clks->decon[CLK_ID_ACLK] = decon_clocks_table[i][CLK_ID_ACLK];
		clks->decon[CLK_ID_PCLK] = decon_clocks_table[i][CLK_ID_PCLK];
		clks->decon[CLK_ID_DPLL] = decon_clocks_table[i][CLK_ID_DPLL];
		break;
	}

	decon_dbg("%s: VCLK %ld ECLK %ld ACLK %ld PCLK %ld DPLL %ld\n",
		__func__,
		clks->decon[CLK_ID_VCLK],
		clks->decon[CLK_ID_ECLK],
		clks->decon[CLK_ID_ACLK],
		clks->decon[CLK_ID_PCLK],
		clks->decon[CLK_ID_DPLL]);
}
